from typing import Protocol, TypeVar

IN = TypeVar("IN", contravariant=True)
OUT = TypeVar("OUT", covariant=True)


class AsyncCallable(Protocol[IN, OUT]):
    async def __call__(self, input_: IN) -> OUT:
        ...



import asyncio
from clubbi_utils.common.iter_as_chunks import iter_as_chunks
from typing import AsyncIterator, Awaitable, Callable, Iterable, List, TypeVar, overload, Literal, Union

OUTPUT = TypeVar("OUTPUT")
INPUT = TypeVar("INPUT")


@overload
def batched_gather(
    f: Callable[[INPUT], Awaitable[OUTPUT]],
    collection: Iterable[INPUT],
    number_of_workers: int,
    return_exceptions: Literal[True],
) -> AsyncIterator[List[Union[Exception, OUTPUT]]]:
    pass


@overload
def batched_gather(
    f: Callable[[INPUT], Awaitable[OUTPUT]],
    collection: Iterable[INPUT],
    number_of_workers: int,
    return_exceptions: bool = False,
) -> AsyncIterator[List[OUTPUT]]:
    pass


async def batched_gather(  # type: ignore[misc]
    f: Callable[[INPUT], Awaitable[OUTPUT]],
    collection: Iterable[INPUT],
    number_of_workers: int,
    return_exceptions: bool = False,
) -> AsyncIterator[List[Union[Exception, OUTPUT]]]:
    for batch in iter_as_chunks(collection, number_of_workers):
        coros = (f(input_) for input_ in batch)
        chunk = list(await asyncio.gather(*coros, return_exceptions=return_exceptions))
        yield chunk



from functools import lru_cache, wraps
from os import getenv
from typing import Optional, Any, Callable

from aiobotocore.client import AioBaseClient
from aiobotocore.config import AioConfig
from aiobotocore.session import AioSession
from aiobotocore.session import get_session
from pydantic import BaseSettings

from clubbi_utils.aws.local_mocks.s3_object_storage_local_mock import S3ObjectStorageLocalMock
from clubbi_utils.aws.s3_object_storage import S3ObjectStorage


@lru_cache
def _get_session() -> AioSession:
    return get_session()


class AioBotocoreSettings(BaseSettings):
    boto_region_name: Optional[str] = None
    boto_read_timeout: int = 60
    boto_max_retry_attempts: int = 5


def with_aws_client(
    name: str,
    region_name: Optional[str] = None,
    read_timeout: Optional[int] = None,
    max_retry_attempts: Optional[int] = None,
) -> Callable:
    def wrap(f):
        async def wrapper(*args, **kwargs):
            settings_kwargs = dict(
                boto_region_name=region_name,
                boto_read_timeout=read_timeout,
                boto_max_retry_attempts=max_retry_attempts,
            )
            settings_kwargs_filtered = {k: v for k, v in settings_kwargs.items() if v is not None}
            settings = AioBotocoreSettings(**settings_kwargs_filtered)

            session = _get_session()
            config = AioConfig(
                read_timeout=settings.boto_read_timeout,
                retries=dict(max_attempts=settings.boto_max_retry_attempts),
            )

            session_kwargs = dict(region_name=settings.boto_region_name) if settings.boto_region_name != None else {}
            async with session.create_client(name, config=config, **session_kwargs) as client:
                return await f(client, *args, **kwargs)

        return wrapper

    return wrap


@lru_cache
def _get_cached_local_mock(
    bucket_name: str,
    key_prefix: str,
) -> S3ObjectStorageLocalMock:
    return S3ObjectStorageLocalMock(
        name=bucket_name,
        key_prefix=key_prefix,
    )


def with_object_storage(
    bucket_name: str,
    key_prefix: str = "",
    region_name: str = "us-east-1",
    env: Optional[str] = None,
) -> Callable:
    """Injects a positional parameter which is a S3ObjectStorage if env is staging or production
    otherwise returns a cached S3ObjectStorageLocalMock."""
    env = env or getenv("ENVIRONMENT")

    def prod_staging_decorator(f: Callable) -> Callable:
        @wraps(f)
        @with_aws_client("s3", region_name=region_name)
        async def wrapper(client: AioBaseClient, *args: Any, **kwargs: Any) -> Any:
            object_storage = S3ObjectStorage(
                client=client,
                bucket=bucket_name,
                key_prefix=key_prefix,
            )
            return await f(object_storage, *args, **kwargs)

        return wrapper

    def local_decorator(f: Callable) -> Callable:
        @wraps(f)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            object_storage = _get_cached_local_mock(bucket_name, key_prefix)
            return await f(object_storage, *args, **kwargs)

        return wrapper

    if env in {"production", "staging"}:
        return prod_staging_decorator
    return local_decorator



from typing import Union, Any, Callable, Set

from pydantic import BaseModel

try:
    import orjson
    from orjson import dumps as _dumps, loads as _loads, OPT_INDENT_2, OPT_SORT_KEYS, OPT_NON_STR_KEYS
    from decimal import Decimal

    def _serialize_default(obj: Any) -> Any:
        if isinstance(obj, Decimal):
            return str(obj)
        if isinstance(obj, BaseModel):
            return obj.dict()
        if isinstance(obj, Set):
            return list(obj)
        raise TypeError

    def dumps(
        data: Any, /, pretty: bool = False, sort_keys: bool = False, non_str_keys: bool = True, default: Callable[[Any], Any] = _serialize_default
    ) -> str:
        option = 0
        if pretty:
            option |= OPT_INDENT_2
        if sort_keys:
            option |= OPT_SORT_KEYS
        if non_str_keys:
            option |= OPT_NON_STR_KEYS

        return _dumps(data, option=option, default=default).decode()

    def loads(data: str) -> Union[str, dict, list, int, float]:
        return _loads(data)

except ImportError:
    from json import dumps as default_dumps, loads as default_loads
    from pydantic.json import pydantic_encoder
    from decimal import Decimal

    def _default(obj: Any) -> Any:
        if isinstance(obj, Decimal):
            return str(obj)
        return pydantic_encoder(obj)

    def dumps(
        data: Any, /, pretty: bool = False, sort_keys: bool = False, non_str_keys: bool = True, default: Callable[[Any], Any] = _default
    ) -> str:
        return default_dumps(data, default=default, sort_keys=sort_keys, indent=2 if pretty else None)

    def loads(data: str) -> Union[str, dict, list, int, float]:
        return default_loads(data)



from typing import Callable, cast
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import Selectable

from clubbi_utils.logging import logger


def print_literal_query(q: Selectable) -> None:
    dialect = cast(Callable, postgresql.dialect)()
    logger.info(q.compile(dialect=dialect, compile_kwargs=dict(literal_binds=True)))



from pydantic import BaseModel
from typing import Iterable, Type, TypeVar
from io import StringIO
from csv import DictWriter


T = TypeVar("T", bound=BaseModel)


def base_model_collection_to_csv(
    model_type: Type[T],
    collection: Iterable[T],
    delimiter: str = ",",
) -> str:
    """Converts a colection of BaseModels to a CSV string,
    ## Example:
    ```
    from pydantic import BaseModel
    class Foo(BaseModel):
        n: int
        s: str

    collection = [Foo(n=1, s="a"), Foo(n=2, s="b")]
    print(base_model_collection_to_csv(Foo, collection))
    # n,s
    # 1,a
    # 2,b
    ```
    Args:
        model_type (Type[BaseModel]): Specific class of the datamodel
        collection (Iterable[BaseModel]): An iterable of the basemodel specified in `model_type`
        delimiter (str, optional): Csv delimiter. Defaults to ",".

    Returns:
        str: CSV as string
    """
    if not issubclass(model_type, BaseModel):
        raise ValueError("model_type parameter must be a subclass of BaseModel")
    f = StringIO()
    writer = DictWriter(
        f=f,
        fieldnames=model_type.schema().get("properties").keys(),  # type: ignore
        delimiter=delimiter,
    )
    writer.writeheader()
    for obj in collection:
        if not isinstance(obj, model_type):
            raise ValueError(f"Each objects in collection must be of type {model_type}")
        writer.writerow(obj.dict())
    return f.getvalue()



from asyncio import StreamReader
from dataclasses import dataclass
from typing import Dict, Generic, Any, Type
from aiobotocore.client import AioBaseClient
from clubbi_utils.async_callable import AsyncCallable, IN, OUT
from pydantic import parse_obj_as, parse_raw_as
from pydantic.error_wrappers import ValidationError

from clubbi_utils import json


@dataclass
class LambdaRaisedAnExceptionError(RuntimeError):
    function_name: str
    payload: Dict

    def __str__(self) -> str:
        return f"Lambda '{self.function_name}' raised the following exception: {json.dumps(self.payload, pretty=True)}"


@dataclass
class InvalidInputType(ValueError):
    expected: Type

    def __str__(self) -> str:
        return f"Invalid input type, expected: {self.expected}"


@dataclass
class FailedToParseOutputError(RuntimeError):
    function_name: str
    output: Any

    def __str__(self) -> str:
        return f"AwsLambdaInvoker '{self.function_name}' couldn't parse the following output: {json.dumps(self.output, pretty=True)}"


class AwsLambdaInvoker(AsyncCallable, Generic[IN, OUT]):
    """Generic lambda invoker class, with an input type `IN` and output type `OUT` passed as parameters, the input type is
    checked at runtime and the output type is parsed if `OUT` is a dataclass, or is a subclass of `pydantic.BaseModel`.

    Args:
        aws_lambda_client (AioBaseClient): aiobotocore base client
        function_name (str): lambda function name
        input_type (Type[IN]): Input type to be checked at runtime
        output_type (Type[OUT]): Output type to be parsed
    """

    def __init__(
        self,
        aws_lambda_client: AioBaseClient,
        function_name: str,
        input_type: Type[IN],
        output_type: Type[OUT],
    ):
        self._client = aws_lambda_client
        self._function_name = function_name
        self._input_type = input_type
        self._output_type = output_type

    def __check_for_errors(self, payload: Dict) -> None:
        if "errorMessage" in payload:
            raise LambdaRaisedAnExceptionError(
                function_name=self._function_name,
                payload=payload,
            )

    def __parse_response_body(self, decoded_body: Any) -> OUT:
        try:
            return parse_obj_as(self._output_type, decoded_body)
        except ValidationError as e:
            raise FailedToParseOutputError(
                function_name=self._function_name,
                output=decoded_body,
            ) from e

    def __check_input_type(self, payload_str: str) -> None:
        try:
            parse_raw_as(self._input_type, payload_str)
        except ValidationError as e:
            raise InvalidInputType(expected=self._input_type) from e
    
    async def __call__(self, input_: IN) -> OUT:
        payload_str = json.dumps(input_)
        self.__check_input_type(payload_str)
        payload_bytes = payload_str.encode()
        lambda_response = await self._client.invoke(
            FunctionName=self._function_name,
            Payload=payload_bytes,
        )

        payload_reader: StreamReader = lambda_response["Payload"]
        raw_body = await payload_reader.read()
        decoded_body = json.loads(raw_body.decode())

        if isinstance(decoded_body, dict):
            self.__check_for_errors(decoded_body)
        
        return self.__parse_response_body(decoded_body)



from collections.abc import Iterable
from typing import AsyncContextManager, Optional, Union

from sqlalchemy import select, BigInteger, values, column, literal
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncConnection
from sqlalchemy.sql.functions import func

_lock = func.pg_advisory_lock
_unlock = func.pg_advisory_unlock


class _PostgresAdvisoryLockManager:

    def __init__(self, engine: AsyncEngine, key: Union[int, Iterable[int]]):
        self._engine = engine
        self._conn: Optional[AsyncConnection] = None

        if isinstance(key, (list, tuple, set)):
            subquery = values(column("k", BigInteger), name="t").data(
                [(literal(key_, BigInteger),) for key_ in set(key)]
            )
            self._lock = select(_lock(subquery.c.k)).select_from(subquery)
            self._unlock = select(_unlock(subquery.c.k)).select_from(subquery)
        else:
            key_ = literal(key, BigInteger)
            self._lock = select(_lock(key_))
            self._unlock = select(_unlock(key_))

    async def __aenter__(self) -> None:
        self._conn = await self._engine.connect()
        await self._conn.execute(self._lock)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._conn:
            await self._conn.execute(self._unlock)
            await self._conn.close()


class PostgresAdvisoryLocker:
    def __init__(self, engine: AsyncEngine):
        self._engine = engine

    def acquire(self, key: Union[int, Iterable[int]]) -> AsyncContextManager[None]:
        return _PostgresAdvisoryLockManager(self._engine, key)



from sqlalchemy.dialects import mysql
from sqlalchemy.sql import Selectable

from clubbi_utils.logging import logger


def print_literal_query(q: Selectable) -> None:
    logger.info(q.compile(dialect=mysql.dialect(), compile_kwargs=dict(literal_binds=True)))



from typing import Optional, Dict, TypeVar, Generic

from aiobotocore.client import AioBaseClient

from clubbi_utils.json import dumps
from clubbi_utils.logging import logger

# docs: https://aws.amazon.com/sns/faqs/
MAXIMUM_MESSAGE_LENGTH = 256_000


class MaximumMessageLengthError(RuntimeError):
    pass


E = TypeVar("E")


class SNSPublisher(Generic[E]):
    def __init__(self, client: AioBaseClient, topic_arn: str):
        self.topic_arn = topic_arn
        self._client = client

    async def publish(
        self, message: E, attributes: Optional[Dict[str, str]] = None, message_group_id: Optional[str] = None
    ) -> None:
        json_body = dumps(message)
        if len(json_body) > MAXIMUM_MESSAGE_LENGTH:
            err_msg = f"The following message is larger than the maximum message size({MAXIMUM_MESSAGE_LENGTH}): '{json_body}'"
            raise MaximumMessageLengthError(err_msg)
        message_attributes = {}
        if attributes:
            message_attributes = {
                k: dict(
                    DataType="String",
                    StringValue=v,
                )
                for k, v in attributes.items()
            }

        publish_kwargs = dict(TopicArn=self.topic_arn, Message=json_body, MessageAttributes=message_attributes)
        if message_group_id is not None:
            publish_kwargs["MessageGroupId"] = message_group_id
        response = await self._client.publish(**publish_kwargs)
        logger.info({"MessageId": response})



from typing import Callable, Optional, AsyncContextManager, AsyncIterator

from pydantic import BaseSettings
from sqlalchemy.engine import URL
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine, AsyncSession
from sqlalchemy.orm import sessionmaker

SessionMaker = Callable[[], AsyncContextManager[AsyncSession]]

from typing import Any, Callable, Coroutine, TypeVar

F = TypeVar(
    "F",
    bound=Callable[..., Coroutine[Any, Any, Any]],
)
T = TypeVar("T")


class SqlAlchemyConfig(BaseSettings):
    user: str
    database: str
    host: str
    port: int
    password: str
    pool_min_size: int = 5
    pool_max_size: int = 10
    pool_timeout: int = 30  # default query timeout
    pool_recycle: int = -1
    echo: bool = False
    driver_name: str

    class Config:
        env_prefix = "sqlalchemy_"

    def to_sqlalchemy_uri(self) -> URL:
        return URL.create(
            drivername=self.driver_name,
            username=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.database,
        )

    def create_engine(self) -> AsyncEngine:
        """engine is meant to be used alongside Sqlalchemy-Core"""

        engine = create_async_engine(
            self.to_sqlalchemy_uri(),
            echo=self.echo,
            pool_size=self.pool_min_size,
            max_overflow=self.pool_max_size - self.pool_min_size,
            pool_timeout=self.pool_timeout,
            pool_recycle=self.pool_recycle,
        )

        return engine

    async def init_engine(self) -> AsyncIterator[AsyncEngine]:
        """this method is meant to be used alongside sqllachemy-core sand frameworks such as:
        - aiohttp
        - FastAPI,
        - dependency-injector
        """
        engine = self.create_engine()
        try:
            yield engine
        finally:
            await engine.dispose()

    async def init_session_maker(self) -> AsyncIterator[SessionMaker]:
        """this method is meant to be used alongside sqlalchemy-orm and frameworks such as:
        - aiohttp
        - FastAPI,
        - dependency-injector
        """
        engine = self.create_engine()
        try:
            yield create_sqlalchemy_session_maker(engine=engine)
        finally:
            await engine.dispose()

    def with_engine(self, f: F) -> Callable[..., Coroutine[Any, Any, T]]:
        """Mostly useful along side lambda and sqlalchemy-core"""

        async def wrapper(*args: Any, **kwargs: dict) -> Any:
            engine = self.create_engine()
            try:
                r = await f(engine, *args, **kwargs)
                return r
            finally:
                await engine.dispose()

        return wrapper

    def with_session_maker(self, f: F) -> Callable[..., Coroutine[Any, Any, T]]:
        """Mostly useful along side lambda and sqlalchemy-orm"""

        async def wrapper(*args: Any, **kwargs: dict) -> Any:
            engine = self.create_engine()
            session_maker = create_sqlalchemy_session_maker(engine=engine)
            try:
                r = await f(session_maker, *args, **kwargs)
                return r
            finally:
                await engine.dispose()

        return wrapper


def create_sqlalchemy_session_maker(
    *, engine: Optional[AsyncEngine], config: Optional[SqlAlchemyConfig] = None, expire_on_commit: bool = False
) -> SessionMaker:
    assert not (engine is not None and config is not None)
    if config:
        engine = config.create_engine()
    elif engine is None:
        engine = SqlAlchemyConfig().create_engine()
    return sessionmaker(engine, expire_on_commit=expire_on_commit, class_=AsyncSession)



import abc
from typing import Generic, TypeVar

T = TypeVar("T")


class AsyncSender(abc.ABC, Generic[T]):
    @abc.abstractmethod
    async def send(self, data: T) -> None:
        pass


AsyncSenderStr = AsyncSender[str]



# Docs https://docs.aws.amazon.com/lambda/latest/dg/services-apigateway.html
from typing import Dict, TypedDict


class AwsLambdaHttpRequestEvent(TypedDict):
    body: str
    headers: Dict[str, str]


class AwsLambdaHttpResponse(TypedDict):
    statusCode: int
    body: str
    headers: Dict[str, str]
